#include "gui_window.h"
#include <exception>

/**
 * Thrown when the window is closed before drawing is completed.
 */
class closed_exception : public std::exception {
public:
    const char *what() const noexcept override { return "Window closed"; }
};

std::map<HWND, Gui_Window *> Gui_Window::open_windows;
ATOM Gui_Window::window_class = (ATOM)nullptr;

LRESULT CALLBACK Gui_Window::window_proc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
    switch (uMsg) {
    case WM_DESTROY:
        PostQuitMessage(0);
        return 0;
    case WM_PAINT: {
        auto window = open_windows.find(hWnd);
        if (window != open_windows.end()) {
            window->second->draw();
            return 0;
        }
    }
    }

    return DefWindowProc(hWnd, uMsg, wParam, lParam);
}

Gui_Window::Gui_Window() {
    if (!window_class) {
        WNDCLASS wc = {};
        wc.lpfnWndProc = &Gui_Window::window_proc;
        wc.hInstance = GetModuleHandle(NULL);
        wc.lpszClassName = "Points";
        wc.hCursor = LoadCursor(NULL, IDC_ARROW);
        window_class = RegisterClass(&wc);
    }

    RECT pos = { 0, 0, window_size + border*2, window_size + border*2 };
    AdjustWindowRect(&pos, WS_OVERLAPPEDWINDOW, FALSE);
    window = CreateWindowEx(0, (LPCSTR)window_class, "Points", WS_OVERLAPPEDWINDOW,
                            CW_USEDEFAULT, CW_USEDEFAULT,
                            pos.right - pos.left, pos.bottom - pos.top,
                            NULL, NULL, GetModuleHandle(NULL), NULL);
    open_windows[window] = this;
    ShowWindow(window, SW_SHOWDEFAULT);

    background = CreateSolidBrush(RGB(0xFF, 0xFF, 0xFF));
    linePen = CreatePen(PS_SOLID, 0, RGB(0x00, 0x00, 0xFF));

    check_events();
}

Gui_Window::~Gui_Window() {
    DeleteObject(background);
    DeleteObject(linePen);
}

Point Gui_Window::transform(Point p) {
    double x = double(p.x) * window_size / max_coord + border;
    double y = double(max_coord - p.y) * window_size / max_coord + border;
    return Point(int(x), int(y));
}

void Gui_Window::draw(const std::vector<Point> &points) {
    for (Point p : points) {
        this->points.push_back(transform(p));
    }

    InvalidateRect(window, NULL, FALSE);
    check_events();
}

void Gui_Window::draw(Point from, Point to) {
    lines.push_back(transform(from));
    lines.push_back(transform(to));

    HDC dc = GetDC(window);
    HGDIOBJ old = SelectObject(dc, linePen);
    int i = lines.size() - 2;
    MoveToEx(dc, lines[i].x, lines[i].y, NULL);
    LineTo(dc, lines[i+1].x, lines[i+1].y);
    SelectObject(dc, old);
    ReleaseDC(window, dc);

    check_events();
}

void Gui_Window::run() {
    MSG msg = {};
    while (GetMessage(&msg, NULL, 0, 0)) {
        TranslateMessage(&msg);
        DispatchMessage(&msg);
    }
}

void Gui_Window::draw() {
    RECT rect;
    GetClientRect(window, &rect);

    PAINTSTRUCT ps;
    HDC hdc = BeginPaint(window, &ps);

    FillRect(hdc, &rect, background);

    for (Point p : points) {
        SetPixel(hdc, p.x, p.y, RGB(0x0, 0x0, 0x0));
    }

    HGDIOBJ old = SelectObject(hdc, linePen);
    for (size_t i = 0; i < lines.size(); i += 2) {
        MoveToEx(hdc, lines[i].x, lines[i].y, NULL);
        LineTo(hdc, lines[i+1].x, lines[i+1].y);
    }
    SelectObject(hdc, old);

    EndPaint(window, &ps);
}

void Gui_Window::check_events() {
    MSG msg = {};
    while (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) {
        if (msg.message == WM_QUIT)
            throw closed_exception();

        TranslateMessage(&msg);
        DispatchMessage(&msg);
    }
}
